/***************************************************************************\
  "inputbuffer.h" - Input Buffer Module (Phase A)
  Revision 3; 10 Nov 2002 - 14 Nov 2002
  Copyright (C) 2002  Kevin P. Albrecht

  References:
    MCD 2.1.1; 2.1.1.1; 2.1.12.1 (Input Buffer Stack)
\***************************************************************************/

#ifndef INPUTBUFFER_H
#define INPUTBUFFER_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

#define INPUTBUFFER_LINE_LEN 256
#define INPUTBUFFER_EOF      0x7F

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdio.h>

// Local //
#include "k_string.h"
#include "error.h"
#include "global.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// External //
void ib_init ();
void ib_deinit ();
void ib_include (char *aFilename);
char ib_get ();
void ib_seek ();

// Internal //
char mPopLine ();

typedef struct Buffer_tag
{
  FILE *file; // input file
  char *filename; // filename
  char text[INPUTBUFFER_LINE_LEN]; // text buffer
  int pos; // current position in mText
  char curr; // current character
  int line; // current line number in file

  struct Buffer_tag *next; // pointer to the next buffer in the stack
} Buffer;

void mBufferFree (Buffer *buffer);
Buffer *mBufferStackTop;

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// ib_init
//// Initialize the Input Buffer Module
void ib_init ()
{
  mBufferStackTop = NULL;
}

//// ib_deinit
//// Deinitialize the input buffer functions
void ib_deinit ()
{
  Buffer *temp;

  // free the buffer stack //
  while (mBufferStackTop)
  {
    temp = mBufferStackTop->next;
    mBufferFree (mBufferStackTop);
    mBufferStackTop = temp;
  }
}

//// ib_include
//// Include a new file
void ib_include (char *aFilename)
{
  Buffer *tempnew;

  // if there is already a file open, set the line number //
  if (mBufferStackTop)
  {
    mBufferStackTop->line = gnCurrentLine;
  }

  // reset the current line and file information //
  gnCurrentLine = 1;
  k_copy (gsCurrentFilename, aFilename);

  // Create new buffer //
  tempnew = (Buffer *) malloc (sizeof (Buffer));

  // Assign values to new buffer //
  tempnew->filename = k_new (strlen (aFilename));
  k_copy (tempnew->filename, aFilename);
  tempnew->file = fopen (tempnew->filename, "r");
  if (!tempnew->file)
  {
    err (false, EP_FILE_OPEN_FAILED, tempnew->filename);
    erre (EFF_TERMINATE);
  }
  k_copy (tempnew->text, "");
  tempnew->curr = '\0';
  tempnew->pos = 0;
  
  // Put the new buffer on top of the stack //
  tempnew->next = mBufferStackTop;
  mBufferStackTop = tempnew;

  tempnew->curr = mPopLine ();
}

//// ib_get
//// Return the current character
char ib_get ()
{
  if (!mBufferStackTop) return 0;

  return mBufferStackTop->curr;
}

//// ib_seek
//// Set the current character to the next character
void ib_seek ()
{
  if (!mBufferStackTop) return;

  if (mBufferStackTop->curr == INPUTBUFFER_EOF) // currently at end of file
  {
    return;
  }
  else if (mBufferStackTop->curr == 0) // currently at end of line
  {
    mBufferStackTop->curr = mPopLine ();
  }
  else
  {
    mBufferStackTop->curr = mBufferStackTop->text[++mBufferStackTop->pos];
  }
}

//// mPopLine
//// Put the next line from the file into the buffer.
//// Then return the line's first character.
char mPopLine ()
{
  char *sTemp; // test to see if 'fgets' was successful
  Buffer *bufferTemp;

  if (!mBufferStackTop) return '\0';

  mBufferStackTop->pos = 0;

  if (feof (mBufferStackTop->file))
  {
    // This is only the end of the currently included file //
    if (mBufferStackTop->next)
    {
      bufferTemp = mBufferStackTop;
      mBufferStackTop = mBufferStackTop->next;
      mBufferFree (bufferTemp);

      // set the current line and file inforamtion //
      gnCurrentLine = mBufferStackTop->line - 1;
      k_copy (gsCurrentFilename, mBufferStackTop->filename);

      return mBufferStackTop->text[mBufferStackTop->pos];
    }
    // This is the end the entire program //
    else
    {
      return INPUTBUFFER_EOF;
    }
  }
  else
  {
    sTemp = fgets (mBufferStackTop->text, INPUTBUFFER_LINE_LEN, mBufferStackTop->file);

    //BUGFIX (14 July 2002 -- 0061):
    //  If the file was not terminated by a newline, the last char was erased
    if (mBufferStackTop->text[strlen(mBufferStackTop->text) - 1] < '\n')
    {
      mBufferStackTop->text[strlen(mBufferStackTop->text) - 1] = '\0';
    }
    //gnCurrentLine++;

    if (mBufferStackTop->text && sTemp) return mBufferStackTop->text[mBufferStackTop->pos];
	  else return 0;
  }
}

//// mBufferFree
//// Free the buffer
void mBufferFree (Buffer *buffer)
{
  if (buffer)
  {
    k_del (buffer->filename);
    fclose (buffer->file);
    free (buffer);
  }
}

#endif
